/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _KNOWN_F_DA_HPP
#define _KNOWN_F_DA_HPP

// a mechanism for reading in known features and (if available) known
// data associations, e.g., from a simulation

#include "slam.hpp"
#include "logfile.hpp"
#include <string.h>
#include <ctype.h>

// this inherits from logfile_t, but we don't actually want to use
// most of its members --- just the file I/O ones --- so we'll make it
// a private inheritance
template <class Feature>
class log_known_f_da : private logfile_t
{
public:
  bool open(const char *filename) { return logfile_t::open(filename); }
  bool reopen() { return logfile_t::reopen(); }
  bool eof() const { return logfile_t::eof(); }

  // read the features (and data associations) from a single frame
  bool read_frame();

  std::vector<Feature> & features() { return cur_features; }
  const std::vector<Feature> & features() const { return cur_features; }
  std::vector<uint32_t> & associations() { return cur_assoc; }
  const std::vector<uint32_t> & associations() const { return cur_assoc; }

protected:
  std::vector<Feature> cur_features; // feature parameters
  std::vector<uint32_t> cur_assoc;   // data associations

  virtual datatype_t read_datum() { return ODOMETRY; }
};

// write a frame of data to a known features/associations log file
template <class Feature>
void write_f_da_to_file
  (const std::vector<Feature> &features,
   const std::vector<uint32_t> &assoc,
   FILE *file);


//////////////////////////////////////////////////////////////////////
// implementation details
//////////////////////////////////////////////////////////////////////

/*
  File format:

  startframe frame-number
  F feature-parameters
  A landmark-id
  endframe
  ... repeats
*/

template <class Feature>
void write_features_to_file
  (const std::vector<Feature> &features, FILE *file)
{
  fprintf(file, "startframe\n");
  for(uint32_t i = 0; i < features.size(); ++i) {
    fprintf(file, "F ");
    features[i].write_to_file(file);
  }
  fprintf(file, "endframe\n");
}

template <class Feature>
void write_features_and_da_to_file
  (const std::vector<Feature> &features,
   const std::vector<uint32_t> &assoc,
   FILE *file)
{
  assert(features.size() == assoc.size());
  fprintf(file, "startframe\n");
  for(uint32_t i = 0; i < features.size(); ++i) {
    fprintf(file, "F ");
    features[i].write_to_file(file);
    fprintf(file, "A %u\n", assoc[i]);
  }
  fprintf(file, "endframe\n");
}

void feature_from_tokens(const std::vector<char *> &tokens, extent_feature &f)
{
  f.r() = atof(tokens[1]);
  f.t() = atof(tokens[2]);
  f.umin = atof(tokens[3]);
  f.umax = atof(tokens[4]);
  f.cov[0][0] = atof(tokens[5]);
  f.cov[0][1] = f.cov[1][0] = atof(tokens[6]);
  f.cov[1][1] = atof(tokens[7]);
}

void feature_from_tokens(const std::vector<char *> &tokens, point_feature &f)
{
  f.x() = atof(tokens[1]);
  f.y() = atof(tokens[2]);
  f.cov[0][0] = atof(tokens[3]);
  f.cov[0][1] = f.cov[1][0] = atof(tokens[4]);
  f.cov[1][1] = atof(tokens[5]);
}

template <class Feature>
bool log_known_f_da<Feature>::read_frame()
{
  cur_features.clear();
  cur_assoc.clear();

  char *line = read_line();
  if(!line)
    return false;

  tokens.clear();
  tokenize(line, tokens);
  assert(tokens.size() == 1 && !strcmp(tokens[0], "startframe"));

  Feature f;

  while(1) {
    line = read_line();
    if(!line)
      return false;

    tokens.clear();
    tokenize(line, tokens);

    if(!strcmp(tokens[0], "endframe"))
      break;

    // read in a feature/association
    if(tokens[0][0] == 'F') {
      feature_from_tokens(tokens, f);
      cur_features.push_back(f);
    } else if(tokens[0][0] == 'A')
      cur_assoc.push_back(strtoul(tokens[1], 0, 10));
    else
      return false;
  }

  return true;
}

#endif // _KNOWN_F_DA_HPP
