/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// more or less straightforward RBPF SLAM (aka FastSLAM 1.0)
// implementation, plus a few small improvements

#include "rbpf.hpp"


//////////////////////////////////////////////////////////////////////
// rbpf implementation
//////////////////////////////////////////////////////////////////////

slam_rbpf::slam_rbpf(const pose_t &start_pose)
  : generic_slam_rbpf<particle_t>()
{
  for(uint32_t i = 0; i < param::N; ++i) {
    particles[i].pose = start_pose;
    particles[i].history.push_back(start_pose);
    particles[i].w = 1.0/param::N;
  }
}

void slam_rbpf::add_odometry(const pose_t &delta, const pose_t &ref_frame)
{
  pose_t mu;
  cov3_t cov;
  // project all the particles forward
  for(uint32_t i = 0; i < particles.size(); ++i) {
    particle_t &p = particles[i];
    cov = cov_epsilon; // motion_model is additive wrt cov
    motion_model(delta, ref_frame, p.pose, mu, cov);
    p.pose = gaussian(mu, cov); // sample new pose
    p.history.push_back(p.pose); // update pose history
  }
}

void slam_rbpf::add_features
  (const std::vector<feature_type> &features,
   const std::vector<uint32_t> &associations,
   const pose_t &ref_frame)
{
  // either we know all associations or none
  assert(associations.size() == 0 || associations.size() == features.size());

  double W = 0, mahalanobis_dist = 0;
  uint32_t i, f;
  std::vector<map_type::iterator> in_range;
  map_type::iterator assoc;
  bool add_new = false;
  feature_type feature;

  // iterate over all particles
  for(i = 0; i < particles.size(); ++i) {
    particle_t &p = particles[i];

    // find in-range landmarks
    if(associations.size() == 0)
      p.map.in_range(p.pose, param::max_range, in_range);

    // process one feature at a time
    for(f = 0; f < features.size(); ++f) {
      // transform into the frame of this particle
      feature_to_particle_frame(features[f], ref_frame, feature, p.pose);

      // do data association
      if(associations.size() != 0) { // known data association
	assoc = p.map.find_id(associations[f]);
	if(assoc != 0) {
	  mahalanobis_dist = mahalanobis_distance(assoc->landmark, feature);
	  add_new = false;
	} else
	  add_new = true;
      } else if(in_range.size() > 0) { // unknown data association
	data_association(feature, in_range, assoc, mahalanobis_dist);
	add_new = (mahalanobis_dist >= param::da_ml_threshold); // are we too far away?

	std::cout << "ft: " << feature << std::endl << feature.cov << std::endl
		  << "lm: " << assoc->landmark << std::endl << assoc->landmark.cov << std::endl
		  << "md: " << mahalanobis_dist << std::endl << std::endl;

      } else
	add_new = true;

      if(!add_new) { // found an association
	p.w *= equality_likelihood_from_mahalanobis(assoc->landmark, feature, mahalanobis_dist);
	assoc = p.map.modify(assoc); // make sure we can touch this landmark
	assoc->landmark = merge(assoc->landmark, feature); // update landmark
	++p.obs_count[assoc];
      } else { // add new landmark to map
	if(associations.size() != 0)
	  p.map.insert(feature, associations[f]); // known landmark id
	else
	  p.map.insert(feature);
	p.w *= exp(-0.5*param::da_ml_threshold*param::da_ml_threshold); // update importance weight
      }
    }

    if(isnan(p.w))
      p.w = 0;

    W += p.w; // update total importance weight over all particles

    // remove landmarks we should have seen but didn't
    remove_spurious_landmarks(p.pose, p.map, in_range, all_points, p.obs_count);
  }

  // this is kind of a catastrophic failure, but do what we can
  if(W == 0) {
    for(i = 0; i < particles.size(); ++i)
      particles[i].w = 1.0;
    W = double(particles.size());
  }

  // normalize importance weights and compute "best" particle
  best = 0;
  for(i = 0; i < particles.size(); ++i) {
    particles[i].w /= W;
    if(particles[i].w > particles[best].w)
      best = i;
  }

  Neff = neff(particles);

  // check if we need to resample, and if so, do it
  if(Neff < param::resample_neff_percent * particles.size())
    resample(particles);

  if(param::gui) {
    // for rendering later
    frame = ref_frame;
    cur_features = features;
  }

  do_analysis();
}
