/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef _RBPF_HPP
#define _RBPF_HPP

// generic RBPF SLAM stuff and class definitions for different SLAM
// algorithms

#include "slam.hpp"
#include "resample.hpp"
#include "neff.hpp"
#include "sample_distribution.hpp"
#include <GL/glut.h>


// FIXME: maybe combine this and sample_distribution into an
// "analysis.hpp" file?
// below
template <class P>
void count_unique_particles_at_each_timestep
  (const std::vector<P> &particles, const char *prefix);

// basic particle
struct particle_t
{
  pose_t pose;                // current pose
  pose_history_t history;     // history of all poses
  map_type map;               // map of this particle
  std::vector<uint32_t> n;    // current data association buffer
  double w;                   // importance weight
  std::map<map_type::iterator, int32_t> obs_count;
  particle_t() {}
  particle_t(const particle_t &p)
    : pose(p.pose), history(p.history), map(p.map), n(p.n), w(p.w), obs_count(p.obs_count) {}
  map_type & get_map() { return map; }
  const map_type & get_const_map() const { return map; }
  std::vector<uint32_t> & get_associations() { return n; }
  const std::vector<uint32_t> & get_const_associations() const { return n; }
};

// particle with fixed-lag maps
struct lag_particle_t
{
  pose_t pose;                // current pose
  pose_history_t history;     // history of all poses (for rendering)

  // fixed-lag data (t-L:t)
  std::list<pose_t>   poses;    // poses
  std::list<map_type> maps;     // maps
  std::list< std::vector<uint32_t> > n; // data association buffers

  double w;                   // importance weight

  std::map<map_type::iterator, int32_t> obs_count;

  lag_particle_t() {}
  lag_particle_t(const lag_particle_t &p)
    : pose(p.pose), history(p.history),
      poses(p.poses), maps(p.maps), n(p.n), w(p.w), obs_count(p.obs_count) {}
  map_type & get_map() { return maps.front(); }
  const map_type & get_const_map() const { return maps.front(); }
  std::vector<uint32_t> & get_associations() { return n.front(); }
  const std::vector<uint32_t> & get_const_associations() const { return n.front(); }
};


// implements some generic stuff that most particle filters can reuse;
// can't actually be used as a slam algorithm itself since it doesn't
// implement add_odometry or add_features; implementation details are
// below
template <class Particle>
class generic_slam_rbpf
{
public:
  generic_slam_rbpf();
  virtual ~generic_slam_rbpf() {}
  void add_scan(const multiscan_t &points);
  virtual void add_odometry(const pose_t &delta, const pose_t &ref_frame) = 0;
  virtual void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t()) = 0;

  // this should be called at the end of add_features for all the
  // filters
  void do_analysis() const { sample_distribution(particles, param::sample_dist_log); }

  void render();
  void write_to_file(const char *prefix) const;
  double get_neff() const { return Neff; }
protected:
  std::vector<Particle> particles;
  uint32_t best; // best particle
  double Neff;

  // most recent scan and extracted features
  pose_t frame;
  cartesian_scan_t all_points;
  std::vector<feature_type> cur_features;
};


//
// standard particle filter (aka fs1.0)
// in rbpf.cpp
//
class slam_rbpf : public generic_slam_rbpf<particle_t>
{
public:
  slam_rbpf(const pose_t &start_pose = pose_t());
  void add_odometry(const pose_t &delta, const pose_t &ref_frame);
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t());
};

//
// rao-blackwellized constraint filter based on fs1.0 (see wafr06
// paper)
// in rbpf_constrained.cpp
//
struct constrained_particle_t : public particle_t
{
  std::map<map_type::iterator, feature_type> accum;  // meas. accumulator
  std::map<map_type::iterator, feature_type> backup; // pre-constrained state
  std::map<map_type::iterator, uint32_t> L;          // superlandmark id
  uint32_t L_max;
  std::map<uint32_t, uint32_t> L_size;               // superlandmark sizes
};

// generic constraint type so we can compile without any constraints
template <class Feature>
struct unconstrained
{
  typedef bool param_type;
  bool draw_constraint_params
    (const Feature &f1, const Feature &f2,
     param_type &c) const { return false; }
  param_type default_c() const { return false; }
  uint32_t constrained_params_size() const { return 0; }
  void get_constrained_params(const Feature &f, Vecd &mu, Matd &cov) const {}
  void h(const param_type &c, Vecd &mu) const {}
  void start_ml_mean(Vecd &mu) const {}
  void add_to_ml_mean(Vecd &mu, const Vecd &umu, const Matd &ucov) const {}
  void end_ml_mean(Vecd &mu) const {}
  void conditionalize(Feature &f, const param_type &c, const Vecd &ml_mean) const {}
};

// rectilinearity constraints on line/segment landmarks
struct rectilinearity
{
  typedef double param_type;
  bool draw_constraint_params
    (const extent_feature &f1, const extent_feature &f2,
     param_type &c) const;
  uint32_t constrained_params_size() const { return 1; }
  param_type default_c() const { return 0; }
  void get_constrained_params(const extent_feature &f, Vecd &mu, Matd &cov) const;
  void h(const param_type &c, Vecd &mu) const;
  void start_ml_mean(Vecd &mu) const;
  void add_to_ml_mean(Vecd &mu, const Vecd &umu, const Matd &ucov) const;
  void end_ml_mean(Vecd &mu) const;
  void conditionalize(extent_feature &f, const param_type &c, const Vecd &ml_mean) const;
};

class slam_rbpf_constrained : public generic_slam_rbpf<constrained_particle_t>
{
public:
  slam_rbpf_constrained(const pose_t &start_pose = pose_t());
  void add_odometry(const pose_t &delta, const pose_t &ref_frame);
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t());
protected:
  void update
    (constrained_particle_t &p, map_type::iterator lm, feature_type &feat);
void slam_rbpf_constrained::insert
  (constrained_particle_t &p, std::vector<map_type::iterator> &landmarks,
   const feature_type &feat, uint32_t id = 0);
};

// make sure you set this appropriately
typedef unconstrained<feature_type> constraint_type;
//typedef rectilinearity constraint_type;


//
// particle filter with improved proposal (aka fs2.0)
// in rbpf_improved.cpp
//
point_feature add_pose_uncertainty
  (const point_feature &feature, const pose_t &mu, const cov3_t &cov);
void improve_proposal
  (pose_t &mu, cov3_t &cov,
   const point_feature &feature, const point_feature &landmark);
extent_feature add_pose_uncertainty
  (const extent_feature &feature, const pose_t &mu, const cov3_t &cov);
void improve_proposal
  (pose_t &mu, cov3_t &cov,
   const extent_feature &feature, const extent_feature &landmark);

// update M with measurements z (wrt frame z_frame) given
// correspondences (landmark ids) n, the robot's pose x, and the
// current likelihood (which will be updated by multiplying it with
// the measurement likelihoods).  since some of the features may be
// new, the convention i'm using is that a correspondence value of 0
// mean "new feature" (so, no landmark in the map should ever have id
// 0!).  this is for fs2.0-like mappers so the fs1.0 implementation
// above doesn't use it
void update_map_improved
  (map_type &m, const std::vector<feature_type> &z,
   const std::vector<uint32_t> &n, const pose_t &x,
   double &weight,
   const pose_t *mu = 0, const cov3_t *cov = 0); // distribution from which x was drawn, optional

// the member functions of this class are templated and implemented in
// this file so they can be used by rbfp_roughening and rbpf_block
// with a different type of particle
template <class Particle>
class slam_rbpf_improved : public generic_slam_rbpf<Particle>
{
public:
  slam_rbpf_improved(const pose_t &start_pose = pose_t())
    : generic_slam_rbpf<Particle>(), last_delta(0,0,0), last_ref_frame(0,0,0),
      new_odometry(true)
  {
    for(uint32_t i = 0; i < param::N; ++i) {
      generic_slam_rbpf<Particle>::particles[i].pose = start_pose;
      generic_slam_rbpf<Particle>::particles[i].history.push_back(start_pose);
      generic_slam_rbpf<Particle>::particles[i].w = 1.0/param::N;
    }
    generic_slam_rbpf<Particle>::do_analysis(); // compute start pose statistics
  }
  void add_odometry(const pose_t &delta, const pose_t &ref_frame);
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t())
  {
    internal_add_features(features, associations, ref_frame);
    generic_slam_rbpf<Particle>::do_analysis();
  }
protected:
  pose_t last_delta, last_ref_frame;
  bool new_odometry;

  void internal_add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t());
};


//
// particle filter with fixed-lag roughening (aka resample-move)
// in rbpf_roughening.cpp
//
struct fixed_lag_data
{
  pose_t u, u_ref_frame;              // control
  std::vector<feature_type> z;        // measurements
  pose_t z_ref_frame;
};

template <class List, class Data>
void fixed_lag_push_front(List &lst, Data &d)
{
#if 0 // FIXME remove
  lst.push_front(d);
#endif
  lst.insert(lst.begin(), d);
  if(lst.size() > param::lag_time+1)
    lst.pop_back();
}

class slam_rbpf_roughening : public slam_rbpf_improved<lag_particle_t>
{
public:
  slam_rbpf_roughening(const pose_t &start_pose = pose_t());
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t());
protected:
  // fixed-lag data that isn't particle-specific (t-L:t)
  std::list<fixed_lag_data> lag_data;
  void fixed_lag_sample(lag_particle_t &p, uint32_t k);
};


//
// particle filter that uses a block proposal to draw poses over a
// fixed lag time
// in rbpf_block.cpp
//
typedef struct { pose_t mu; cov3_t cov; } pose_model_t; // convenience

#if 0
class slam_rbpf_block : public slam_rbpf_roughening
{
public:
  slam_rbpf_block(const pose_t &start_pose = pose_t())
    : slam_rbpf_roughening(start_pose) {}
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t());
protected:
  double filter_forward
    (const fixed_lag_data &d, const pose_model_t &last,
     pose_model_t &cur, map_type &tmp_map, std::vector<uint32_t> &associations);
};
#endif
//#if 0
struct block_particle_t
{
  pose_t pose;                // current pose
  pose_history_t history;     // history of all poses (for rendering)

  // fixed-lag data (t-L:t)
  std::vector<pose_t>   poses;    // poses
  std::vector<map_type> maps;     // maps
  std::vector< std::vector<uint32_t> > n; // data association buffers

  double w;                   // importance weight

  std::map<map_type::iterator, int32_t> obs_count;

  block_particle_t() {}
  block_particle_t(const block_particle_t &p)
    : pose(p.pose), history(p.history),
      poses(p.poses), maps(p.maps), n(p.n), w(p.w), obs_count(p.obs_count) {}
  map_type & get_map() { return maps[0]; }
  const map_type & get_const_map() const { return maps[0]; }
  std::vector<uint32_t> & get_associations() { return n[0]; }
  const std::vector<uint32_t> & get_const_associations() const { return n[0]; }
};
class slam_rbpf_block : public slam_rbpf_improved<block_particle_t>
{
public:
  slam_rbpf_block(const pose_t &start_pose = pose_t());
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t());
protected:
  // fixed-lag data that isn't particle-specific (t-L:t)
  std::vector<fixed_lag_data> lag_data;
  double filter_forward
    (const fixed_lag_data &d, const pose_model_t &last,
     pose_model_t &cur, map_type &tmp_map, std::vector<uint32_t> &associations);
};
//#endif

//////////////////////////////////////////////////////////////////////
// templated utilities
//////////////////////////////////////////////////////////////////////

template <class P>
void count_unique_particles_at_each_timestep
  (const std::vector<P> &particles, const char *prefix)
{
  FILE *file = open_file(prefix, "unique");
  if(!file) return;

  typedef pose_history_t::node node;
  std::map<const node *, bool> cur, next;
  for(uint32_t i = 0; i < particles.size(); ++i) {
    assert(particles[i].history.length == particles[0].history.length);
    next[particles[i].history.head] = true;
  }
  std::map<const node *, bool>::iterator mi;
  do {
    cur = next;
    fprintf(file, "%d\n", cur.size());
    next.clear();
    for(mi = cur.begin(); mi != cur.end(); ++mi) {
      if(mi->first->next == 0) break;
      next[mi->first->next] = true;
    }
  } while(mi == cur.end());
  fclose(file);
}

//////////////////////////////////////////////////////////////////////
// slam_rbpf_improved member functions (parts of the slam
// implementation, e.g., the improved proposal, are implemented in
// rbpf_improved.cpp since they didn't need to be templated
//////////////////////////////////////////////////////////////////////

template <class Particle>
void slam_rbpf_improved<Particle>::add_odometry
  (const pose_t &delta, const pose_t &ref_frame)
{
  if(!new_odometry) {
    // add_odometry has already been called since the last call to
    // add_features.  apply a prediction step and store the resulting
    // pose update in the lag data.  it will be treated as a step with
    // no observed features (which it is).
    static std::vector<feature_type> empty_features(0);
    static std::vector<uint32_t>     empty_associations(0);
    static pose_t                    dummy_ref_frame;
    add_features(empty_features, empty_associations, dummy_ref_frame);
  }
  // store the odometry for use in computing the improved proposal
  // distribution
  last_delta = delta;
  last_ref_frame = ref_frame;
  new_odometry = false;
}

template <class Particle>
void slam_rbpf_improved<Particle>::internal_add_features
  (const std::vector<feature_type> &features,
   const std::vector<uint32_t> &associations,
   const pose_t &ref_frame)
{
  // either we know all associations or none
  assert(associations.size() == 0 || associations.size() == features.size());

  pose_t mu;
  cov3_t cov;

  double W = 0, mahalanobis_dist;
  uint32_t i, f;
  std::vector<map_type::iterator> in_range;
  map_type::iterator assoc;
  feature_type feat;

  generic_slam_rbpf<Particle>::cur_features.resize(features.size());

  // inflate maximum range a bit since we haven't updated the pose yet
  double my_max_range = param::max_range +
    ::sqrt(last_delta.x()*last_delta.x()+last_delta.y()*last_delta.y());

  // iterate over all particles
  for(i = 0; i < generic_slam_rbpf<Particle>::particles.size(); ++i) {
    Particle &p = generic_slam_rbpf<Particle>::particles[i];

    // find in-range landmarks
    if(associations.size() != features.size())
      p.get_map().in_range(p.pose, my_max_range, in_range);


    //////////////////////////////////////////////////////////////////
    // compute improved proposal distribution (using motion model and
    // all the observed features)

    // apply the motion model to get the basic odometry proposal
    cov = cov_epsilon;
    motion_model(last_delta, last_ref_frame, p.pose, mu, cov);

    for(f = 0; f < generic_slam_rbpf<Particle>::cur_features.size(); ++f) {
      // incorporate measurements in the proposal
      feature_to_particle_frame(features[f], ref_frame, feat, mu);
      if(associations.size()) // already have a data association
	assoc = p.get_map().find_id(associations[f]);
      else{
	if(in_range.size()) {
	  // do data association (inflate uncertainty by current proposal)
	  data_association(add_pose_uncertainty(feat, mu, cov), in_range, assoc, mahalanobis_dist);
	  if(mahalanobis_dist >= param::da_ml_threshold)
	    assoc = 0;
	} else assoc = 0;
      }
      if(assoc)
	improve_proposal(mu, cov, feat, assoc->landmark);
    }

    // done computing improved proposal
    //////////////////////////////////////////////////////////////////

    p.pose = gaussian(mu, cov); // sample new pose
    p.history.push_back(p.pose); // update pose history

    // transform features to the sampled frame of this particle
    for(f = 0; f < features.size(); ++f)
      feature_to_particle_frame
	(features[f], ref_frame,
	 generic_slam_rbpf<Particle>::cur_features[f], p.pose);

    p.get_associations().resize(generic_slam_rbpf<Particle>::cur_features.size());

    // do data association conditioned on the sampled pose
    if(associations.size() != features.size()) {
      for(f = 0; f < generic_slam_rbpf<Particle>::cur_features.size(); ++f) {
	if(in_range.size())
	  data_association
	    (generic_slam_rbpf<Particle>::cur_features[f], in_range,
	     assoc, mahalanobis_dist);
	if(!in_range.size() || mahalanobis_dist >= param::da_ml_threshold)
	  p.get_associations()[f] = 0; // no association, we'll add a new landmark
	else {
	  ++p.obs_count[assoc]; // FIXME: need to do this for known DA as well!
	  p.get_associations()[f] = assoc->id;
	}
      }
    } else
      p.get_associations() = associations; // copy known data associations

    // update map with the current observations and correspondences
    update_map_improved
      (p.get_map(), generic_slam_rbpf<Particle>::cur_features,
       p.get_associations(), p.pose, p.w, &mu, &cov);

    if(isnan(p.w))
      p.w = 0;

    W += p.w; // update total importance weight over all particles

    // remove landmarks we should have seen but didn't
    remove_spurious_landmarks
      (p.pose, p.get_map(), in_range,
       generic_slam_rbpf<Particle>::all_points, p.obs_count);
  }

  // this is kind of a catastrophic failure, but do what we can
  if(W == 0) {
    for(i = 0; i < generic_slam_rbpf<Particle>::particles.size(); ++i)
      generic_slam_rbpf<Particle>::particles[i].w = 1.0;
    W = double(generic_slam_rbpf<Particle>::particles.size());
  }

  // normalize importance weights and compute "best" particle
  generic_slam_rbpf<Particle>::best = 0;
  for(i = 0; i < generic_slam_rbpf<Particle>::particles.size(); ++i) {
    generic_slam_rbpf<Particle>::particles[i].w /= W;
    if(generic_slam_rbpf<Particle>::particles[i].w >
       generic_slam_rbpf<Particle>::particles[generic_slam_rbpf<Particle>::best].w)
      generic_slam_rbpf<Particle>::best = i;
  }

  generic_slam_rbpf<Particle>::Neff = neff(generic_slam_rbpf<Particle>::particles);

  // check if we need to resample, and if so, do it
  if(generic_slam_rbpf<Particle>::Neff <
     param::resample_neff_percent * generic_slam_rbpf<Particle>::particles.size())
    resample(generic_slam_rbpf<Particle>::particles);

  // we've now done a SLAM update up to the most recent control input
  new_odometry = true;

  if(param::gui) {
    // for rendering later
    generic_slam_rbpf<Particle>::frame = ref_frame;
    generic_slam_rbpf<Particle>::cur_features = features;
  }
}

//////////////////////////////////////////////////////////////////////
// implementation details for generic_slam_rbpf
//////////////////////////////////////////////////////////////////////

template <class P>
generic_slam_rbpf<P>::generic_slam_rbpf()
  : particles(param::N), best(0) {}

template <class P>
void generic_slam_rbpf<P>::add_scan(const multiscan_t &points)
{
  frame = points.back().frame;
  agglomerate_multiscan(points, all_points);
  all_points.radial_sort(frame);

  std::vector<feature_type> ext_features;
  extract_features(all_points, ext_features);
  if(ext_features.size() == 0)
    return;

  std::vector<uint32_t> dummy(0); // we don't know associations
  add_features(ext_features, dummy, frame);
}

template <class P>
void generic_slam_rbpf<P>::render()
{
  uint32_t i;

  const P &bp = particles[best];

  // draw the FOV underneath everything else
  glColor3f(0.3,0.1,0.1);
  glBegin(GL_LINE_STRIP);
  glVertex2f(bp.pose.x(), bp.pose.y());
  for(double a = bp.pose.t()+param::scan_min; a <= bp.pose.t()+param::scan_max; a += 5.0*M_PI/180)
    glVertex2f(bp.pose.x()+param::max_range*cos(a), bp.pose.y()+param::max_range*sin(a));
  glVertex2f(bp.pose.x(), bp.pose.y());
  glEnd();

  // first, all pose histories
  glLineWidth(1);
  glColor3f(0, 0.5, 0);
  for(i = 0; i < particles.size(); ++i)
    particles[i].history.render(200); // FIXME change this?

  // next, the best map ...
  glColor3f(0,0,1);
  bp.get_const_map().render();

  // ... and the best pose history ...
  glLineWidth(1);
  glColor3f(1, 0, 1);
  bp.history.render();

  // ... and the best current pose ...
  glPointSize(8);
  glColor3f(0, 0.7, 0);
  glBegin(GL_POINTS);
  glVertex2f(bp.pose.x(), bp.pose.y());
  glEnd();
  glBegin(GL_LINES);
  glVertex2f(bp.pose.x(), bp.pose.y());
  glVertex2f(bp.pose.x()+cos(bp.pose.t()), bp.pose.y()+sin(bp.pose.t()));
  glEnd();

  // ... and the scan points ...
  //    transform_features_and_scan(cur_features, all_points, frame, bp.pose);
  glColor3f(1, 0, 0);
  glPointSize(2);
  all_points.render();

  // ... and finally, features from the current scan
  glColor3f(1,0,0);
  feature_type feat;
  for(i = 0; i < cur_features.size(); ++i){
    feature_type &feature = cur_features[i]; 
    feature_to_particle_frame(feature, frame, feat, bp.pose);
    feat.render(true); // renders covariance
#ifndef _LINE_FEATURES // FIXME
    glBegin(GL_LINES);
    glVertex2d(bp.pose.x(), bp.pose.y());
    glVertex2d(feat.x(), feat.y());
    glEnd();
#endif
  }
}

template <class P>
void generic_slam_rbpf<P>::write_to_file(const char *prefix) const
{
  particles[best].get_const_map().write_to_file(prefix);
  particles[best].history.write_to_file(prefix);
  full_sample_distribution(particles, param::full_sample_dist_log);
  count_unique_particles_at_each_timestep(particles, prefix);
}


//////////////////////////////////////////////////////////////////////
// implementation details for slam_interface
//////////////////////////////////////////////////////////////////////

// a generic class that can be used by the controlling program to
// access the proper slam algorithm depending on the command line
// arguments; basically these functions are all switch statements
class slam_interface
{
public:
#define pick_and_run(call) \
  switch(param::slam_alg) { \
  case FS1: return fs1.call; break; \
  case FS2: return fs2.call; break; \
  case ROUGH: return rough.call; break; \
  case BLOCK: return block.call; break; \
  case CONSTRAINED: return constrained.call; break; \
  }

  void initialize(const pose_t &start_pose = pose_t())
  {
    fs1 = slam_rbpf(start_pose);
    fs2 = slam_rbpf_improved<particle_t>(start_pose);
    rough = slam_rbpf_roughening(start_pose);
    block = slam_rbpf_block(start_pose);
    constrained = slam_rbpf_constrained(start_pose);
    reset_analysis_logs(); // because we instantiate several filters
    do_analysis(); // redo initial analysis
  }
  void add_scan(const multiscan_t &points) { pick_and_run(add_scan(points)); }
  void add_odometry(const pose_t &delta, const pose_t &ref_frame)
    { pick_and_run(add_odometry(delta, ref_frame)); }
  void add_features
    (const std::vector<feature_type> &features,
     const std::vector<uint32_t> &associations,
     const pose_t &ref_frame = pose_t())
    { pick_and_run(add_features(features, associations, ref_frame)); }
  void do_analysis() const { pick_and_run(do_analysis()); }
  void render() { pick_and_run(render()); }
  void write_to_file(const char *prefix) const { pick_and_run(write_to_file(prefix)); }
  double get_neff() const { pick_and_run(get_neff()); return 0; }
protected:
  slam_rbpf fs1;
  slam_rbpf_improved<particle_t> fs2;
  slam_rbpf_roughening rough;
  slam_rbpf_block block;
  slam_rbpf_constrained constrained;
};

#endif // _RBPF_HPP
