/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// RBPF SLAM with improved proposal distribution (aka FastSLAM 2.0)

// this is actually mainly supporting code, the main particle filter
// implementation is in rbpf.hpp because it is templated

#include "rbpf.hpp"

// convert a point feature to a range-bearing measurement with the
// appropriate uncertainty (which INCLUDES pose uncertainty (cov))
void point_to_range_bearing(const pose_t &p, const cov3_t &cov, const point_feature &f,
			    double &r, double &a, cov2_t &c)
{
  double dx = f.x() - p.x();
  double dy = f.y() - p.y();
  r = ::sqrt(dx*dx + dy*dy);
  a = atan2(dy,dx) - p.t();

  // compute range-bearing uncertainty as Prb = JrPrJr' + JfPfJf'
  // where Jr = [[-dx/r,-dy/r,0],[dy/r^2,-dx/r^2,-1]] and Jf =
  // [[dx/r,dy/r],[-dy/r^2,dx/r^2]]
  Matd Jr(2,3, -dx/r, -dy/r, 0.0,
	       dy/(r*r), -dx/(r*r), -1.0);
  Mat2d Jf(dx/r, dy/r, -dy/(r*r), dx/(r*r));
  c = mat2_from_matd(Jr*cov*trans(Jr) + Jf*f.cov*trans(Jf));
}

point_feature add_pose_uncertainty
  (const point_feature &feature, const pose_t &mu, const cov3_t &cov)
{
  // this kind of sucks -- because we assume throughout the code that
  // measurements are in feature parameter space, we need to go
  // through a couple transformations here to add pose uncertainty to
  // the feature uncertainty

  point_feature out(feature);

  double r, a;
  cov2_t c;
  point_to_range_bearing(mu, cov, feature, r, a, c);

  // now turn this back into feature uncertainty: Pff = JffPrbJff'
  const double cosa = cos(a), sina = sin(a);
  static Mat2d J;
  J[0][0] = cosa; J[0][1] = -r*sina; J[1][0] = sina; J[1][1] = r*cosa;
  out.cov = mat2_from_matd(J*c*trans(J));

  return out;
}

void improve_proposal
  (pose_t &mu, cov3_t &cov,
   const point_feature &feature, const point_feature &landmark)
{
  // compute range-bearing uncertainty as Prb = JfPfJf' + JfPlJf' with
  // Jf as above
  double rf, rl, bf, bl;
  cov2_t Prbf, Prbl;
  point_to_range_bearing(mu, cov, feature, rf, bf, Prbf);
  point_to_range_bearing(mu, cov, landmark, rl, bl, Prbl);
  cov2_t invPrb(inv(Prbf + Prbl));

  // ugly since it's already computed in point_to_range_bearing
  double dx = landmark.x() - mu.x();
  double dy = landmark.y() - mu.y();

  // merge gaussians to improve proposal with:
  //  nu = range-bearing(feature) - range-bearing(landmark)
  //  P = inv(Jr'inv(Prb)Jr + inv(P))
  //  mu = mu + P*Jr'*inv(Prb)*nu
  Matd Jr(2,3, -dx/rl, -dy/rl, 0.0,
	       dy/(rl*rl), -dx/(rl*rl), -1.0);
  Vec2d nu(rf-rl, angular_distance(bf,bl));
  cov = mat3_from_matd(inv(trans(Jr)*invPrb*Jr + inv(cov)));
  mu = vec3_from_vecd(mu + cov*trans(Jr)*invPrb*nu);
}

extent_feature add_pose_uncertainty
  (const extent_feature &feature, const pose_t &mu, const cov3_t &cov)
{
  extent_feature out(feature);

  // FIXME: verify this (same as below except for feature)
  static Matd Jr(2,3, 0,0,0, 0,0,1);
  Jr[0][0] = cos(feature.t());
  Jr[0][1] = sin(feature.t());

  out.cov = mat2_from_matd(Jr*cov*trans(Jr) + out.cov);

  return out;
}

void improve_proposal
  (pose_t &mu, cov3_t &cov,
   const extent_feature &feature, const extent_feature &landmark)
{
  // merge gaussians to improve proposal with:
  //  nu = feature - landmark
  //  P = inv(Jr'inv(Pfl)Jr + inv(P))
  //  mu = mu + P*Jr'*inv(Pfl)*nu

  // FIXME: verify this (same as above except for landmark)
  static Matd Jr(2,3, 0,0,0, 0,0,1);
  Jr[0][0] = cos(landmark.t());
  Jr[0][1] = sin(landmark.t());

  cov2_t invPfl(inv(feature.cov + landmark.cov));
  Vec2d nu(feature.r()-landmark.r(), angular_distance(feature.t(),landmark.t()));
  cov = mat3_from_matd(inv(trans(Jr)*invPfl*Jr + inv(cov)));
  mu = vec3_from_vecd(mu + cov*trans(Jr)*invPfl*nu);
}

// see rbpf.hpp for some explanation
// assumes z are in the frame of x!
void update_map_improved
  (map_type &m, const std::vector<feature_type> &z,
   const std::vector<uint32_t> &n, const pose_t &x,
   double &weight,
   const pose_t *mu, const cov3_t *cov) // distribution from which x was drawn, optional
{
  map_type::iterator assoc;
  bool add_new = false;
  feature_type tmp;
  double mahalanobis_dist = 0;

  assert(n.size() == z.size());

  // now process all the features
  for(uint32_t f = 0; f < z.size(); ++f) {
    const feature_type &feature = z[f];

    assoc = (n[f] == 0) ? 0 : m.find_id(n[f]); // don't look if no correspondence
    if(assoc != 0) {
      if(mu && cov)
	tmp = add_pose_uncertainty(feature, *mu, *cov); // inflate covariance by motion model
      else tmp = feature; // ugly
      mahalanobis_dist = mahalanobis_distance(assoc->landmark, tmp);
      add_new = false;
    } else
      add_new = true;

    if(!add_new) { // found an association
      weight *= equality_likelihood_from_mahalanobis(assoc->landmark, tmp, mahalanobis_dist);
      assoc = m.modify(assoc); // make sure we can touch this landmark
      assoc->landmark = merge(assoc->landmark, feature); // update landmark
    } else { // add new landmark to map
      if(n[f] != 0)
	m.insert(feature, n[f]); // known landmark id
      else
	m.insert(feature); // let the map class generate an id
      weight *= exp(-0.5*param::da_ml_threshold*param::da_ml_threshold); // update importance weight
    }
  }
}
