/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// a simple simulator with point or line features --- outputs ground
// truth trajectory, odometry estimated trajectory, and frame-wise
// features and data associations for use with slam

#include "slam.hpp"
#include "fast_landmark_map.hpp"
#include "options.hpp"
#include "known_f_da.hpp"
#include <GL/glut.h>
#include <algorithm>
#include <fstream>

map_type map;
std::list<point_t> waypoints;
pose_history_t truth, odom;
double workspace_w, workspace_h;
FILE *features_file = 0;
pose_t last_input;

namespace param {
  double max_range;
  double wheel_base;
  double err_range, err_bearing, err_trans, err_rot, err_norm, err_vel, err_steering;
  double min_consecutive_dist_2;
  double timestep;
  double velocity;
  double scan_min, scan_res; // for types.cpp, but unused
  bool gui;
  FILE *sample_dist_log;
  FILE *full_sample_dist_log;
  FILE *neff_log;
  FILE *performance_log;
};

// pick a point uniformly at random in the workspace
void generate(point_feature &f)
{
  f.x() = drand48() * workspace_w - workspace_w/2;
  f.y() = drand48() * workspace_h - workspace_h/2;
}

// pick a line segment in the workspace that doesn't intersect with
// any other line segment and that meets our length requirements
void generate(extent_feature &f)
{
  // FIXME FIXME
}

// generate a random "environment"
void generate(uint32_t n)
{
  feature_type f;
  for(uint32_t i = 0; i < n; ++i) {
    generate(f);
    map.insert(f);
  }
}

// generate the ground truth trajectory; basically borrowed from Tim
// Bailey's Matlab SLAM simulator
void generate_perfect_trajectory_from_waypoints()
{
  truth.clear();
  std::list<point_t>::iterator cwp = waypoints.begin(), nwp = cwp;
  ++nwp;
  // compute an initial heading (just point toward the 2nd waypoint)
  double theta = (nwp == waypoints.end()) ? 0 :
    clamp_angle(atan2(nwp->y() - cwp->y(), nwp->x() - cwp->x()));
  double G = 0; // steering angle
  pose_t p(cwp->x(), cwp->y(), theta);
  while(cwp != waypoints.end()) {
    // have we reached the current waypoint?
    if(squared_distance(*cwp, p) < 1.0) { // FIXME make this a parameter?
      // switch to the next waypoint
      ++cwp;
      if(cwp == waypoints.end())
	break;
    }

    // angle to current waypoint
    double dG = clamp_angle(atan2(cwp->y() - p.y(), cwp->x() - p.x()) - p.t() - G);

    // limit rate of steering angle change
    double max_dG = (20*M_PI/180) * param::timestep; // FIXME: make this a parameter
    if(fabs(dG) > max_dG)
      dG = sign(dG)*max_dG;

    // limit angle
    G = G+dG;
    double max_G = (30*M_PI/180); // FIXME: make this a parameter
    if(fabs(G) > max_G)
      G = sign(G)*max_G;

    // compute perfect motion
    // FIXME: make wheelbase (2.0) a parameter
    p += param::velocity * param::timestep * pose_t(cos(G+p.t()), sin(G+p.t()), sin(G)/2.0);
    truth.push_back(p);
  }
}

void get_observed_features
  (const pose_t &p,
   std::vector<map_t<point_feature>::iterator> &out)
{
  map.in_range(p, param::max_range, out);
}

void get_observed_features
  (const pose_t &p,
   std::vector<map_t<extent_feature>::iterator> &out)
{
  //  map.in_range(p, param::max_range, out);

  // now go through and get rid of features that are occluded
  // fIXME FIXME FIXME
}

void transform_and_add_noise_and_covariance(const pose_t &p, point_feature &f)
{
  double range = distance(f, p);
  double bearing = clamp_angle(atan2(f.y()-p.y(), f.x()-p.x()) - p.t());

  // add noise
  range = gaussian(range, param::err_range);
  bearing = gaussian(bearing, param::err_bearing);

  const double sina = sin(bearing), cosa = cos(bearing);

  f.x() = range*cosa;
  f.y() = range*sina;

  // compute covariance of point from range/bearing
  // assume these are absolute, not a function of the range/bearing
  static Mat2d Pm(param::err_range*param::err_range, 0.0, 0.0, param::err_bearing*param::err_bearing);
  static Mat2d J;
  J[0][0] = cosa; J[0][1] = -range*sina; J[1][0] = sina; J[1][1] = range*cosa;
  f.cov = mat2_from_matd(J*Pm*trans(J));
}

void transform_and_add_noise_and_covariance(const pose_t &p, extent_feature &f)
{
  // FIXME
}

// run the simulation
void simulate()
{
  features_file = fopen("sim.feat", "w");
  std::list<pose_t> gt;
  truth.get_stl_list(gt);

  std::vector<map_type::iterator> observed;
  std::vector<feature_type> features;
  std::vector<uint32_t> associations;

  std::list<pose_t>::iterator i = gt.begin();
  pose_t last_pose = *i, last_odom = *i, delta;
  odom.push_back(*i);

  while(1) {
    // get observations (with noise) and associations
    get_observed_features(*i, observed);
    features.clear();
    associations.clear();
    for(uint32_t j = 0; j < observed.size(); ++j) {
      feature_type f(observed[j]->landmark);
      transform_and_add_noise_and_covariance(*i, f);
      features.push_back(f);
      associations.push_back(observed[j]->id);
    }
    write_features_and_da_to_file(features, associations, features_file);

    last_pose = *i;
    ++i;
    if(i == gt.end())
      break;

    // project odometry forward
    delta = *i - last_pose;

    pose_t mu;
    cov3_t cov = cov_epsilon;
    motion_model(delta, last_pose, last_odom, mu, cov);
    last_odom = gaussian(mu, cov);
    odom.push_back(last_odom);
  }

  fclose(features_file);
}


// a lot of the rest of this is just copied from gui.cpp

int mouse_x, mouse_y;
bool mouse_left = false, mouse_right = false, mouse_middle = false;
uint32_t redraw_freq = 1;
int viewport_w, viewport_h, viewport_w_orig, viewport_h_orig;
double scale = 1, trans_x, trans_y, extra_x, extra_y;

void display()
{
  glClearColor(0.7, 0.7, 0.7, 1);
  glClear(GL_COLOR_BUFFER_BIT);

  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluOrtho2D(scale * (-workspace_w / 2 + trans_x),
	     scale * (workspace_w / 2 + trans_x + extra_x),
	     scale * (-workspace_h / 2 + trans_y),
	     scale * (workspace_h / 2 + trans_y + extra_y));

  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();

  // because occupancy grid data is stored at a different orientation
  glRotatef(-90, 0, 0, 1);

  // axes
  glLineWidth(1);
  glColor3f(0.8,0.8,0.8);
  glBegin(GL_LINES);
  glVertex2d(-100000, 0); glVertex2d(100000, 0);
  glVertex2d(0, -100000); glVertex2d(0, 100000);
  glEnd();

  glPointSize(5);
  glColor3f(0,0,1);
  map.render();
  glColor3f(0,0.5,0);
  truth.render();
  glColor3f(0.5,0,0);
  odom.render();

  // waypoints
  glBegin(GL_POINTS);
  glPointSize(8);
  glColor3f(0,0.8,0);
  for(std::list<point_t>::const_iterator i = waypoints.begin(); i != waypoints.end(); ++i)
    glVertex2f(i->x(), i->y());
  glEnd();

  glutSwapBuffers();
}

void reshape(int width, int height)
{
  glViewport(0, 0, width, height);
  viewport_w = width;
  viewport_h = height;
  extra_x = workspace_w * (float(width) - viewport_w_orig) / float(viewport_w_orig);
  extra_y = workspace_h * (float(height) - viewport_h_orig) / float(viewport_h_orig);
}

void keyboard(uint8_t key, int x, int y)
{
  switch(key) {
  case 'Q':
  case 'q': // quit
    printf("\ngoodbye\n");
    truth.write_to_file("sim-truth");
    odom.write_to_file("sim-odom");
    map.write_to_file("sim");
    exit(0);

  case 'T':
  case 't': // generate ground truth trajectory from waypoints
    truth.clear();
    generate_perfect_trajectory_from_waypoints();
    break;

  case 'S':
  case 's': // run simulation
    odom.clear();
    simulate();
    break;

  case 'G':
  case 'g': // generate a new environment and discard previous data
    generate(options::quickget<uint32_t>("num-features"));
    break;
  }
  glutPostRedisplay();
}

void mouse(int button, int state, int x, int y)
{
  mouse_x = x;
  mouse_y = y;

  if(state == GLUT_UP) {
    if(button == GLUT_LEFT_BUTTON) mouse_left = false;
    else if(button == GLUT_RIGHT_BUTTON) mouse_right = false;
    else if(button == GLUT_MIDDLE_BUTTON) { mouse_middle = false; }
  } else {
    if(button == GLUT_LEFT_BUTTON) mouse_left = true;
    else if(button == GLUT_RIGHT_BUTTON) mouse_right = true;
    else if(button == GLUT_MIDDLE_BUTTON) mouse_middle = true;
  }

  if(button == GLUT_MIDDLE_BUTTON && state == GLUT_DOWN) {
    // add a new waypoint to ground truth trajectory
    GLdouble mview[16], proj[16], wx, wy, wz;
    GLint vport[4];
    glGetDoublev(GL_MODELVIEW_MATRIX, mview);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    glGetIntegerv(GL_VIEWPORT, vport);
    gluUnProject(x, y, 0, mview, proj, vport, &wx, &wy, &wz);
    wx = -wx;

    waypoints.push_back(point_t(wx,wy));
  }

  glutPostRedisplay();
}

void motion(int x, int y)
{
  int dx = x - mouse_x, dy = y - mouse_y;
  if(dx == 0 && dy == 0)
    return;

  if(mouse_middle) { // add to ground truth trajectory
#if 0
    // get world pos of pointer, add to ground truth history if
    // distance from last pose is big enough; this is a little tricky
    // because we don't have a theta for the previous pose until we
    // get the next one

    GLdouble mview[16], proj[16], wx, wy, wz;
    GLint vport[4];
    glGetDoublev(GL_MODELVIEW_MATRIX, mview);
    glGetDoublev(GL_PROJECTION_MATRIX, proj);
    glGetIntegerv(GL_VIEWPORT, vport);
    gluUnProject(x, y, 0, mview, proj, vport, &wx, &wy, &wz);
    wx = -wx;

    // FIXME: this isn't really very nice, honestly
    if(!started) {
      started = true;
      truth.clear();
      last_input = pose_t(wx,wy,0);
    } else {
      double d_2 = (wx-last_input.x)*(wx-last_input.x) + (wy-last_input.y)*(wy-last_input.y);
      if(d_2 >= param::min_consecutive_dist_2) {
	last_input.t = atan2(wy-last_input.y, wx-last_input.x);
	truth.push_back(last_input);
	last_input = pose_t(wx,wy,0);
      }
    }
#endif
  } else if(mouse_left) { // pan
    trans_x -= workspace_w * float(dx) / float(viewport_w_orig);
    trans_y += workspace_h * float(dy) / float(viewport_h_orig);
  } else if(mouse_right) // zoom
    scale *= exp(0.05 * double(dy) + 0.01);

  mouse_x = x;
  mouse_y = y;

  if(mouse_left || mouse_right || mouse_middle)
    glutPostRedisplay();
}

void special(int key, int x, int y)
{
  if(key == GLUT_KEY_HOME) {
    scale = 1.0;
    trans_x = trans_y = 0.0;
  }
  glutPostRedisplay();
}

int parse_command_line(int argc, char **argv)
{
  // set up commandline/configuration file options
  options::add<bool>("help", 0, "Print usage information", 0, false, options::nodump);
  options::set_cf_options("config", "c");
  options::add<std::string>("save-config", 0, "Save configuration file", 0, "", options::nodump);

  options::add<bool>("nogui", 0, "No GUI (quietly simulate with given map/truth)",
		     "Graphical display", false, options::nodump);
  options::add<double>("workspace-width", "w", "Workspace width (m)", "Graphical display", 80);
  options::add<double>("workspace-height", "h", "Workspace height (m)", "Graphical display", 80);
  options::add<int>("gui-width", 0, "GUI width (pixels)", "Graphical display", 800);
  options::add<int>("gui-height", 0, "GUI height (pixels)", "Graphical display", 800);

  options::add<double>("max-range", 0, "Maximum laser range (m)", "Model", 10.0);
  options::add<double>("err-range", 0, "Range error (exact, in m)",
		       "Model", 0.05);
  options::add<double>("err-bearing", 0, "Bearing error (exact, in radians)",
		       "Model", 0.0001);
  options::add<double>("err-translation", 0, "Translation error (percent of odometry measurement)",
		       "Model", 0.12);
  options::add<double>("err-rotation", 0, "Rotation error (percent of odometry measurement)",
		       "Model", 0.24);
  options::add<double>("err-normal", 0, "Normal error (percent of odometry measurement)",
		       "Model", 0.12);

  options::add<std::string>("map", "m", "Map file", "Input files", "");
  options::add<std::string>("truth", "t", "Ground truth trajectory file", "Input files", "");

  options::add<uint32_t>("num-features", "n", "Number of features to generate", "Simulation", 100);
  options::add<double>("min-consecutive-dist", "d", "Minimum L2 distance between consecutive poses (m)",
		       "Simulation", 0.15);

  // parse commandline (and read config file if it is specified)
  int inpidx = options::parse_cmdline(argc, argv);

  if(inpidx < 0) // some kind of error
    exit(1);    // relevant info printed by getopt

  // print usage information
  if(options::quickget<bool>("help") == true) {
    std::cerr << "Usage: " << argv[0] << " [options]" << std::endl;
    options::print_options(std::cout);
    exit(1);
  }

  // save a config file based on these options?
  std::string cfname = options::quickget<std::string>("save-config");
  if(cfname.length() > 0) {
    std::ofstream conf(cfname.c_str());
    if(!conf)
      std::cerr << "Can't write configuration file " << cfname << std::endl;
    else
      options::dump(conf);
  }

  // save options in variables
  param::max_range = options::quickget<double>("max-range");
  param::err_range = options::quickget<double>("err-range");
  param::err_bearing = deg2rad(options::quickget<double>("err-bearing"));
  param::err_trans = options::quickget<double>("err-translation");
  param::err_rot = options::quickget<double>("err-rotation");
  param::err_norm = options::quickget<double>("err-normal");
  param::min_consecutive_dist_2 = options::quickget<double>("min-consecutive-dist");
  param::min_consecutive_dist_2 *= param::min_consecutive_dist_2;
  param::timestep = 0.2; // FIXME: make a commandline option
  param::velocity = 2; // FIXME: make a commandline option
  param::gui = !options::quickget<bool>("nogui");

  return inpidx;
}

int main(int argc, char **argv)
{
  parse_command_line(argc, argv);

  workspace_w = std::max(options::quickget<double>("workspace-width"), 0.1);
  workspace_h = std::max(options::quickget<double>("workspace-height"), 0.1);

  srand48(time(0));

  // load or generate environment
  if(options::quickget<std::string>("map").length() > 0)
    map.read_from_file(options::quickget<std::string>("map").c_str());
  else
    generate(options::quickget<uint32_t>("num-features"));

  // load ground truth if available
  if(options::quickget<std::string>("truth").length() > 0)
    truth.read_from_file(options::quickget<std::string>("truth").c_str());
  else if(param::gui)
    printf("Use the middle mouse button to place waypoints,"
	   " then press 't' to generate trajectory\n");
  else {
    printf("You need to specify a ground truth or create one with the gui\n");
    exit(1);
  }

  if(param::gui)
    printf("Press 's' run the simulation\n");
  else {
    simulate();
    odom.write_to_file("sim-odom");
    if(!(options::quickget<std::string>("map").length() > 0))
      map.write_to_file("sim");
    exit(0);
  }

  // get commandline options that we use for rendering
  viewport_w = viewport_w_orig = std::max(options::quickget<int>("gui-width"), 1);
  viewport_h = viewport_h_orig = std::max(options::quickget<int>("gui-height"), 1);

  // set up renderer
  glutInit(&argc, argv);

  glutInitDisplayMode(GLUT_RGBA | GLUT_DOUBLE);
  glutInitWindowSize(viewport_w, viewport_h);

  glutCreateWindow("simulator");

  glutDisplayFunc(display);
  glutReshapeFunc(reshape);
  glutKeyboardFunc(keyboard);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutSpecialFunc(special);

  glutMainLoop();

  return 0;
}

