/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

#include <stdio.h>
#include "slam.hpp"
#include <algorithm>
#include <sys/time.h>

// this is a program to test a simple k-means clustering algorithm of
// particle trajectories over a short time span (say, 50 poses at 5-10
// Hz).  the goal is to find particles with "similar" trajectories
// (relative to some starting pose) over a given time period.  one
// could do clustering on the full dimensionality of the particles
// over the time period, but this would be prohibitively expensive
// with many particles and more than a few time steps (3 dimensions
// per time step, N particles, T time steps -> clustering of N
// 3T-dimensional points).  so, this calls for some dimensionality
// reduction.  however, we're not really willing to spend much time
// executing, e.g., SVD or some other fairly expensive (but accurate)
// approach.  the result is a sort of "poor man's dimensionality
// reduction" tailored to this particular problem.

// the first thing we'll do is cluster only on the (x,y) values of the
// poses (ignoring the thetas).  one reason for this is that, as
// always, it's not clear how to define "distance" between (x,y,theta)
// poses.  a partial justification is that, under the assumption that
// the robot drives in a direction related to its heading, this
// heading is to some degree encoded in the consecutive relative (x,y)
// values of the poses.

// this doesn't reduce the dimensionality of the problem sufficiently.
// a simple idea is to compute the mean values of (x,y) for all poses
// in the trajectory, and cluster (using k-means) based on these.
// this turns out to actually reasonably approximate a clustering that
// uses all of the individual (x,y)'s.  we can do a little better by
// dividing the trajectory into M "chunks," computing the mean (x,y)'s
// for each chunk of each particle, and clustering in the resulting
// 2M-dimensional space.  our results with this approach indicate that
// using a few (3-5) chunks is nearly the same as clustering on all of
// the data for reasonably short trajectories.

// here is how to generate some statistics/figures for a simple
// forward motion over T time steps with N particles, C clusters, and
// M chunks (all constants defined below):

// if _CHUNK_STATS is not defined, the output will be a file where
// each line is a single particle's number, the assigned cluster id,
// and then the trajectory of that particle.  pass this file to
// plot_clustered.m to make a figure with different-colored
// trajectories for the different clusters.

// define _CHUNK_STATS to compute the clustering for all chunk counts
// between 1 and T, using the same initial clustering assignments, and
// then compute the difference from the clusters resulting from T
// chunks.  this basically shows that a reasonably small number of
// chunks gives results that are more or less the same as with many
// chunks, so we can make do with a few chunks and save on computation
// in clustering.  if you've defined _CHUNK_STATS, you should run the
// program repeatedly with the following bash script:
//
// LIMIT=1000; for ((a=1; a <= LIMIT; a++)) do echo $a; ./test_particle_cluster >>foo.txt; done
//
// this will generate a file that can then be given to
// plot_cluster_diff.m (along with the number of particles) to produce
// a graph showing the mean number of differently clustered particles
// for different chunk sizes.

//#define _CHUNK_STATS
//#define _OUTPUT_ITERATIONS

// parameters
//const uint32_t N = 100, T = 50, C = 5, M = 5;
const uint32_t N = 100, T = 50, C = 5, M = 15;
const pose_t motion(0.2, 0.0, 0.0); // forward 0.2m each time step
//const pose_t motion(0.2, 0.0, 0.05); // forward 0.2m, turn left
// noise model is defined in main()


#ifdef _CHUNK_STATS
uint32_t clust_seed;
#endif

// simple k-means clustering with random initialization
template <class Point>
void k_means(const std::vector<Point> &pts, std::vector<uint32_t> &cluster,
	     uint32_t k, const Point &zero)
{
  const uint32_t &N = pts.size();
  uint32_t i;

  cluster.resize(N);

#ifdef _CHUNK_STATS
  srand48(clust_seed);
#endif

  std::vector<uint32_t> clust_size(k);
  std::vector<Point> clust_mean(k);
  std::map<uint32_t, bool> picked;
  bool done;
  double d_min, d;
  uint32_t c, c_min;

  // assign initial cluster means by picking random points in the set
  // for each, without replacement to guarantee different clusters
  for(c = 0; c < C; ++c) {
    do {
      c_min = (uint32_t)(drand48()*N);
    } while(picked[c_min]); // without replacement
    picked[c_min] = true;
    clust_mean[c] = pts[c_min];
  }

  uint32_t iter_count = 0;
  while(1) {
    ++iter_count;
    done = true;

    // re-assign points to clusters based on distance from the cluster
    // means
    for(i = 0; i < N; ++i) {
      d_min = sqrlen(pts[i] - clust_mean[0]);
      c_min = 0;
      for(c = 1; c < k; ++c) {
	d = sqrlen(pts[i] - clust_mean[c]);
	if(d < d_min) {
	  c_min = c;
	  d_min = d;
	}
      }
      if(c_min != cluster[i])
	done = false; // different than last time, not converged
      cluster[i] = c_min;
    }

    if(done) // no cluster indices changed
      break;

    // reset cluster means/sizes to zero
    for(c = 0; c < k; ++c) {
      clust_size[c] = 0;
      clust_mean[c] = zero;
    }

    // compute new cluster means
    for(i = 0; i < N; ++i) {
      ++clust_size[cluster[i]];
      clust_mean[cluster[i]] += pts[i];
    }
    for(c = 0; c < k; ++c)
      if(clust_size[c])
	clust_mean[c] /= double(clust_size[c]);
  }
#ifdef _OUTPUT_ITERATIONS
  printf("%d %d\n", k, iter_count);
#endif
}

template <class Particle>
void chunked_trajectory_cluster
  (const std::vector<Particle> &particles, std::vector<uint32_t> &cluster,
   uint32_t clusters, uint32_t chunks)
{
  const uint32_t &N = particles.size();
  const Vecd pn_zero(2*chunks, vl_0);

  std::vector<Vecd> mean_xy_c(N, pn_zero);
  uint32_t i, t, c;

  for(i = 0; i < N; ++i) {
    const Particle &p = particles[i];

    pose_t xform = -1 * p[0];

    // compute mean (x,y) for this particle with respect to the
    // initial pose
    for(c = 0, t = 0; t < p.size(); ++t) { // FIXME: change to get data from a real pose hist
      if(t > (c+1)*p.size() / chunks)
	++c; // next chunk
      point_t pt = transform(point_t(p[t].x(),p[t].y()), xform);
      mean_xy_c[i][2*c] += pt.x();
      mean_xy_c[i][2*c+1] += pt.y();
    }
    mean_xy_c[i] /= double(p.size() / chunks);
  }

  k_means(mean_xy_c, cluster, clusters, pn_zero);
}

int main()
{
  // noise model (pretty noisy)
  param::err_trans = 0.25;
  param::err_norm = 0.1;
  param::err_rot = 0.3;

  typedef std::vector<pose_t> phist;
  std::vector<phist> p(N);
  std::vector<uint32_t> cluster(N);

  struct timeval tv;
  gettimeofday(&tv, 0);
  srand48(tv.tv_usec);
#ifdef _CHUNK_STATS
  clust_seed = tv.tv_usec;
#endif

  uint32_t i = 0;

  for(; i < N; ++i)
    p[i].push_back(pose_t(0,0,0));

  pose_t mu;
  cov3_t cov;
  for(uint32_t t = 0; t < T; ++t) {
    for(i = 0; i < N; ++i) {
      cov = cov_epsilon;
      phist &ph = p[i];
      motion_model(motion, pose_t(0,0,0), ph[ph.size()-1], mu, cov);
      ph.push_back(gaussian(mu,cov));
    }
  }

#ifndef _CHUNK_STATS
  // generates output that can be plotted with plot_clustered.m
  chunked_trajectory_cluster(p, cluster, C, M);

  for(i = 0; i < N; ++i) {
    printf("%d %d", i, cluster[i]);
    const phist &ph = p[i];
    for(uint32_t t = 0; t < T; ++t)
      printf(" %g %g %g", ph[t].x(), ph[t].y(), ph[t].t());
    printf("\n");
  }
#endif

#ifdef _CHUNK_STATS
  // generates output that can be plotted with plot_cluster_diff.m
  std::vector< std::vector<uint32_t> > clusterings(T);
  for(uint32_t c = T-1; c > 0; --c) {
    chunked_trajectory_cluster(p, clusterings[c], C, c);
    uint32_t diff = 0;
    for(i = 0; i < N; ++i)
      if(clusterings[c][i] != clusterings[T-1][i])
	++diff;
    printf(" %d", diff);
  }
  printf("\n");
#endif

  return 0;
}


// hack: put this here so we can link
namespace param
{
  // filters
  uint32_t N; // # particles
  uint32_t m; // scans per multiscan
  slam_algorithm slam_alg;
  resample_algorithm resample_alg;
  double resample_neff_percent;
  double resample_alpha; // only used for generalized resampling
  uint32_t lag_time; // for fixed-lag filters

  // model
  pose_t origin;
  double scan_min, scan_max, scan_res;
  uint32_t sparse_points;
  double max_range;
  double err_range, err_bearing;
  double err_trans, err_rot, err_norm;

  // feature extraction
  double le_min_length;
  uint32_t le_min_points;
  double le_merge_threshold;
  double le_converge_threshold;
  double le_tmax;
  double le_cluster_threshold;

  // data association
  double da_ml_threshold;
  uint32_t spurious_discard_count;

  // misc
  bool gui;
  bool known_features;
  FILE *sample_dist_log;
  FILE *full_sample_dist_log;
  FILE *neff_log;
  FILE *performance_log;
};
