/*
  slam3 (c) 2006 Kris Beevers

  This file is part of slam3.

  slam3 is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  slam3 is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
  or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
  License for more details.

  You should have received a copy of the GNU General Public License
  along with slam3; if not, write to the Free Software Foundation,
  Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA
*/

// basic types (pose, scan, etc...)

#include "slam.hpp"
#include <GL/glut.h>
#include <algorithm>

FILE * open_file(const char *prefix, const char *extension, const char *mode)
{
  char *filename = new char[strlen(prefix) + strlen(extension) + 2];
  sprintf(filename, "%s.%s", prefix, extension);
  FILE *file = fopen(filename, mode);
  delete [] filename;
  return file;
}

void reset_analysis_logs()
{
  close_analysis_logs();
  param::sample_dist_log = open_file("output", "mcphist");
  param::full_sample_dist_log = open_file("output", "mcfull");
  param::neff_log = open_file("output", "neff");
  param::performance_log = open_file("output", "perf");
}

void close_analysis_logs()
{
  if(param::sample_dist_log)
    fclose(param::sample_dist_log);
  if(param::full_sample_dist_log)
    fclose(param::full_sample_dist_log);
  if(param::neff_log)
    fclose(param::neff_log);
  if(param::performance_log)
    fclose(param::performance_log);
}

Mat2d mat2_from_matd(const Matd &in)
{
  return Mat2d(in[0][0], in[0][1],
	       in[1][0], in[1][1]);
}
Vec2d vec2_from_vecd(const Vecd &in)
{
  return Vec2d(in[0], in[1]);
}

Mat3d mat3_from_matd(const Matd &in)
{
  return Mat3d(in[0][0], in[0][1], in[0][2],
	       in[1][0], in[1][1], in[1][2],
	       in[2][0], in[2][1], in[2][2]);
}
Vec3d vec3_from_vecd(const Vecd &in)
{
  return Vec3d(in[0], in[1], in[2]);
}

// merge two pose distribution estimates in the ugliest way imaginable
// (thanks, maple!)
void merge(pose_t &p1, cov3_t &c1, const pose_t &p2, const cov3_t &c2)
{
  Mat3d K = c1*inv(c1+c2);
  p1 += K*(p2-p1);
  c1 = inv(inv(c1)+inv(c2));
}


void pose_history_t::destroy()
{
  node *n;
  while((n = head) && n->count == 1) {
    head = head->next;
    delete n;
  }
  if(head)
    --(head->count);
  length = 0;
}

pose_history_t & pose_history_t::operator=(const pose_history_t &h)
{
  destroy();
  head = h.head;
  length = h.length;
  if(h.head)
    ++(h.head->count);
  return *this;
}

void pose_history_t::render(uint32_t how_far_back) const
{
  glLineWidth(1);
  glBegin(GL_LINE_STRIP);
  uint32_t i = 0;
  for(node *n = head; n != 0 && i < how_far_back; n = n->next, ++i)
    glVertex2d(n->x(), n->y());
  glEnd();
}

void pose_history_t::write_to_file(const char *prefix) const
{
  // get a reversed history (start to end instead of end to start)
  std::list<pose_t> lst;
  get_stl_list(lst);
  FILE *file = open_file(prefix, "phist");
  if(file) {
    std::list<pose_t>::iterator i = lst.begin();
    for(; i != lst.end(); ++i)
      fprintf(file, "%g %g %g\n", i->x(), i->y(), rad2deg(i->t()));
    fclose(file);
  }
}

void pose_history_t::read_from_file(const char *file)
{
  clear();
  FILE *f = fopen(file, "r");
  if(f) {
    float x, y, t;
    while(!feof(f)) {
      fscanf(f, "%g %g %g\n", &x, &y, &t);
      push_back(pose_t(x,y,deg2rad(t)));
    }
  }
}

void pose_history_t::get_stl_list(std::list<pose_t> &out) const
{
  out.clear();
  for(node *n = head; n != 0; n = n->next)
    out.insert(out.begin(), *n);
}

point_t transform(const point_t &p, const pose_t &delta)
{
  point_t o = Rot2d(delta.t()) * p;
  o += point_t(delta.x(),delta.y());
  return o;
}

line_t transform(const line_t &l, const pose_t &delta)
{
  // FIXME?
  line_t o;
  o.t() = l.t()+delta.t();
  o.r() = l.r() + delta.x()*cos(o.t()) + delta.y()*sin(o.t());
  return o;
}

void cartesian_scan_t::render() const
{
  glBegin(GL_POINTS);
  //  glBegin(GL_LINES);
  //  glLineWidth(1);
  for(uint32_t i = 0; i < size(); ++i) {
    //    glColor3f(double(i)/size(), 0.0, 1-double(i)/size());
    //    glVertex2d(frame.x, frame.y);
    glVertex2d((*this)[i].x(), (*this)[i].y());
  }
  glEnd();
}

cartesian_scan_t::cartesian_scan_t(const scan_t &in, const pose_t &_frame)
  : frame(_frame)
{
  reserve(in.size());
  double ang = param::scan_min + frame.t();
  for(uint32_t i = 0; i < in.size(); ++i, ang += param::scan_res)
    if(in[i] < param::max_range)
      push_back(point_t(in[i] * cos(ang) + frame.x(), in[i] * sin(ang) + frame.y()));
  // FIXME: also compute covariance?
}

struct radial_comp
{
  const pose_t &p;
  radial_comp(const pose_t _p) : p(_p) {}
  bool operator()(const point_t &p1, const point_t &p2)
  {
    return atan2(p1.y()-p.y(),p1.x()-p.x()) < atan2(p2.y()-p.y(),p2.x()-p.x());
  }
};
void cartesian_scan_t::radial_sort(const pose_t &p)
{
  std::sort(begin(), end(), radial_comp(p));
}

void cartesian_scan_t::transform(const pose_t &delta)
{
  const double cost = cos(delta.t());
  const double sint = sin(delta.t());
  point_t o;
  for(uint32_t i = 0; i < size(); ++i) {
    const point_t &p = (*this)[i];
    o.x() = delta.x() + p.x()*cost - p.y()*sint;
    o.y() = delta.y() + p.x()*sint + p.y()*cost;
    (*this)[i] = o;
  }
  frame += delta;
}

void agglomerate_multiscan(const multiscan_t &ms, cartesian_scan_t &out)
{
  out.clear();
  for(uint32_t i = 0; i < ms.size(); ++i)
    for(uint32_t j = 0; j < ms[i].size(); ++j)
      out.push_back(ms[i][j]);
}


void line_feature::render() const
{
  glBegin(GL_LINES);
  glVertex2d(-100000, r() + 100000*cos(t())/sin(t()));
  glVertex2d(100000, r() - 100000*cos(t())/sin(t()));
  glEnd();
}

double mahalanobis_distance(const line_feature &f1, const line_feature &f2)
{
  line_t nu(f2-f1);
  double d = sqrt(dot(nu*inv(f1.cov+f2.cov),nu));
  if(isnan(d))
    return max_double();
  return d;
}

line_feature merge(const line_feature &f1, const line_feature &f2)
{
  line_feature ret(f1);
  Mat2d K = f1.cov*inv(f1.cov+f2.cov);
  line_t nu(f2.r()-f1.r(), angular_distance(f2.t(), f1.t()));
  ret += K*nu;
  ret.cov = inv(inv(f1.cov)+inv(f2.cov));
  return ret;
}

line_feature transform(const line_feature &f, const pose_t &delta)
{
  line_feature o(transform((line_t)f, delta));
  const double dS = delta.y()*cos(o.t()) - delta.x()*sin(o.t());
  Mat2d B(1, dS, 0, 1);
  o.cov = B * f.cov * trans(B);
  return o;
}


void extent_feature::render(bool render_uncertainty) const
{
  glColor3f(0,0,1);
  point_t s = u_to_point(*this, umin);
  point_t e = u_to_point(*this, umax);
  glBegin(GL_LINES);
  glVertex2d(s.x(), s.y());
  glVertex2d(e.x(), e.y());
  glEnd();

  if(render_uncertainty) {
    // render a rough approximation of uncertainty information

    // version that only renders range uncertainty
    glColor3f(0,0,0);
    extent_feature err(*this);
    err.r() += ::sqrt(cov[0][0]);
    //    err.t += ::sqrt(cov.[1][1]);
    point_t s = u_to_point(err, umin);
    point_t e = u_to_point(err, umax);
    glBegin(GL_LINES);
    glVertex2d(s.x(), s.y());
    glVertex2d(e.x(), e.y());
    glEnd();
    err.r() -= 2*::sqrt(cov[0][0]);
    //    err.t -= 2*::sqrt(cov[1][1]);
    s = u_to_point(err, umin);
    e = u_to_point(err, umax);
    glBegin(GL_LINES);
    glVertex2d(s.x(), s.y());
    glVertex2d(e.x(), e.y());
    glEnd();

#if 0
    // version that goes +/- 1sigma for both range and bearing
    glColor3f(0,0,0);
    line_t err(*this);
    err.r += ::sqrt(cov.rr);
    err.t += ::sqrt(cov.tt);
    double u = project_onto(s, err);
    point_t es = u_to_point(err, u);
    u = project_onto(e, err);
    point_t ee = u_to_point(err, u);
    glBegin(GL_LINES);
    glVertex2d(es.x, es.y);
    glVertex2d(ee.x, ee.y);
    glEnd();
    err = line_t(*this);
    err.r -= ::sqrt(cov.rr);
    err.t -= ::sqrt(cov.tt);
    u = project_onto(s, err);
    es = u_to_point(err, u);
    u = project_onto(e, err);
    ee = u_to_point(err, u);
    glBegin(GL_LINES);
    glVertex2d(es.x, es.y);
    glVertex2d(ee.x, ee.y);
    glEnd();
#endif
  }
}

void extent_feature::write_to_file(FILE *file) const
{
  point_t s(u_to_point(*this, umin));
  point_t e(u_to_point(*this, umax));
  fprintf(file, "%g %g %g %g %g %g %g %g %g %g %g\n",
	  s.x(), s.y(), e.x(), e.y(), r(), t(), umin, umax,
	  cov[0][0], cov[0][1], cov[1][1]);
}

void extent_feature::read_from_file(FILE *file)
{
  float disc, _r, _t, _umin, _umax, _rr, _rt, _tt;
  fscanf(file, "%g %g %g %g %g %g %g %g %g %g %g\n",
	 &disc, &disc, &disc, &disc, &_r, &_t, &_umin, &_umax,
	 &_rr, &_rt, &_tt);
  r() = _r; t() = _t; umin = _umin; umax = _umax;
  cov[0][0] = _rr; cov[0][1] = cov[1][0] = _rt; cov[1][1] = _tt;
}

double squared_length(const extent_feature &f)
{
  point_t p1(u_to_point(f, f.umin)), p2(u_to_point(f, f.umax));
  return squared_distance(p1, p2);
}

double length(const extent_feature &f)
{
  return ::sqrt(squared_length(f));
}

double squared_distance(const point_t &p, const extent_feature &f)
{
  point_t pt;
  double u = project_onto(p, f);
  if(u >= f.umin && u <= f.umax)
    pt = u_to_point(f, u);
  else if(u < f.umin)
    pt = u_to_point(f, f.umin);
  else
    pt = u_to_point(f, f.umax);

  return squared_distance(p, pt);
}

bool overlap(const extent_feature &f1, const extent_feature &f2)
{
  point_t p = u_to_point(f1, f1.umin);
  double u = project_onto(p, f2);
  if(u >= f2.umin && u <= f2.umax)
    return true;
  p = u_to_point(f1, f1.umax);
  u = project_onto(p, f2);
  if(u >= f2.umin && u <= f2.umax)
    return true;
  p = u_to_point(f2, f2.umin);
  u = project_onto(p, f1);
  if(u >= f1.umin && u <= f1.umax)
    return true;
  return false;
}

// internal use (in merge, below)
inline void update_extent(extent_feature &f, const point_t &p)
{
  double u = project_onto(p, f);
  if(u < f.umin) f.umin = u;
  if(u > f.umax) f.umax = u;
}

extent_feature merge(const extent_feature &f1, const extent_feature &f2)
{
  extent_feature ret(merge((line_feature)f1, (line_feature)f2));
  ret.umin = max_double();
  ret.umax = -max_double();
  update_extent(ret, u_to_point(f1, f1.umin));
  update_extent(ret, u_to_point(f1, f1.umax));
  update_extent(ret, u_to_point(f2, f2.umin));
  update_extent(ret, u_to_point(f2, f2.umax));
  return ret;
}

extent_feature transform(const extent_feature &f, const pose_t &delta)
{
  extent_feature o(transform((line_feature)f, delta));
  point_t p(transform(u_to_point(f, f.umin), delta));
  o.umin = project_onto(p, o);
  p = transform(u_to_point(f, f.umax), delta);
  o.umax = project_onto(p, o);
  if(o.umin > o.umax)
    std::swap(o.umin, o.umax);
  return o;
}

void feature_to_particle_frame
  (const extent_feature &in, const pose_t &in_frame,
   extent_feature &out, const pose_t &out_frame)
{
  // FIXME FIXME verify this!!
  pose_t delta(out_frame - in_frame);
  out = transform(in, delta);
}

void draw_ellipse(const point_t &mu, const cov2_t &sigma, double limit = 0.67)
{
  const double rr = sigma[0][0];
  const double rt = sigma[0][1];
  const double tt = sigma[1][1];
  const double k = sqrt(-2 * log(1 - limit));
  const double T = sqrt(rr*rr - 2*rr*tt +
			tt*tt + 4*rt*rt);
  double deg = rr*tt - rt*rt;
  const bool isdeg = fabs(deg) < 1e-5;
  const double major = fabs(isdeg ? rr+tt : sqrt((2*k*k*deg) / (rr+tt-T)));
  const double minor = fabs(isdeg ? 0 : sqrt((2*k*k*deg) / (rr+tt+T)));
  const double theta = atan2(2*rt, rr-tt) / 2;

  const double inc = 0.3141592654; // 1-degree increments

  glPushMatrix();
  glTranslated(mu.x(), mu.y(), 0.0);
  glRotated(theta*180/M_PI, 0, 0, 1);
  glScaled(1.0, minor/major, 1.0);

  glBegin(GL_TRIANGLE_FAN);
  for(deg = 0.0; deg < 2*M_PI; deg += inc)
    glVertex2d(cos(deg) * major, sin(deg) * major);
  glVertex2d(cos(0.0) * major, sin(0.0) * major); // close ellipse
  glEnd();

  glPopMatrix();
}

void point_feature::render(bool render_uncertainty) const
{
  //  if(render_uncertainty) {
  //    glColor3f(0.5,0.5,1);
  //    draw_ellipse(*this, cov);
  draw_ellipse(*this, cov, 0.99);
    //  }

    //  glColor3f(0,0,1);
  glPointSize(3);
  glBegin(GL_POINTS);
  glVertex2f(x(), y());
  glEnd();
}

void point_feature::write_to_file(FILE *file) const
{
  fprintf(file, "%g %g %g %g %g\n", x(), y(), cov[0][0], cov[0][1], cov[1][1]);
}

void point_feature::read_from_file(FILE *file)
{
  float _x, _y, _xx, _xy, _yy;
  fscanf(file, "%g %g %g %g %g\n", &_x, &_y, &_xx, &_xy, &_yy);
  x() = _x; y() = _y; cov[0][0] = _xx; cov[0][1] = cov[1][0] = _xy; cov[1][1] = _yy;
}

double mahalanobis_distance(const point_feature &f1, const point_feature &f2)
{
  point_t nu = f2-f1;
  double d = sqrt(dot(nu*inv(f1.cov+f2.cov),nu));
  if(isnan(d))
    return max_double();
  return d;
}

point_feature merge(const point_feature &f1, const point_feature &f2)
{
  point_feature ret(f1);
  Mat2d K = f1.cov*inv(f1.cov+f2.cov);
  ret += K*(f2-f1);
  ret.cov = inv(inv(f1.cov)+inv(f2.cov));
  return ret;
}

point_feature transform(const point_feature &f, const pose_t &delta)
{
  point_feature o(transform((point_t)f, delta));
  Mat2d J(Rot2d(delta.t()));
  o.cov = J*o.cov*trans(J);
  return o;
}

void feature_to_particle_frame
  (const point_feature &in, const pose_t &in_frame,
   point_feature &out, const pose_t &out_frame)
{
  // FIXME: not really the ideal way to do this (uh, use transform,
  // four lines above?)!  here i'm turning the point feature into a
  // range-bearing measurement, transforming the measurement to the
  // output frame, and computing a new point feature.  kind of silly
  // but it works for now.
  // FIXME: subtract in_frame.t()?
  double a = atan2(in.y()-in_frame.y(), in.x()-in_frame.x()) + out_frame.t();
  double r = distance(in, in_frame);
  const double cosa = cos(a), sina = sin(a);
  out.x() = out_frame.x() + r*cosa;
  out.y() = out_frame.y() + r*sina;
  static Mat2d Pm(param::err_range*param::err_range, 0.0, 0.0, param::err_bearing*param::err_bearing);
  static Mat2d J;
  J[0][0] = cosa; J[0][1] = -r*sina; J[1][0] = sina; J[1][1] = r*cosa;
  out.cov = mat2_from_matd(J*Pm*trans(J));
}
